use crate::*;
use serde::{Deserialize, Serialize};
use std::collections::HashMap;
use std::io::{stdout, Write};
use std::path::PathBuf;
use std::process::{Command, ExitStatus, Stdio};
use std::{fmt, io};

#[derive(Debug, Serialize, Deserialize, PartialEq)]
enum UpdateSteps {
    Fetch,
    Compile,
    Install,
}

#[derive(Debug, Serialize, Deserialize)]
pub struct Updater {
    systems: Vec<System>,
}

#[derive(Debug, Serialize, Deserialize)]
pub struct System {
    //name: String,
    fetch: Option<Cmd>,
    compile: Option<Cmd>,
    install: Cmd,
    // deps or rDeps : Tree
    // exclusive_with : List
}

#[derive(Debug, Serialize, Deserialize, Clone)]
struct Cmd {
    exe: String,
    params: Vec<String>,
    current_dir: Option<PathBuf>,
    env: HashMap<String, String>,
}

#[derive(Debug, Serialize, Deserialize)]
pub(crate) struct ActualCmd {
    exe: String,
    params: Vec<String>,
    current_dir: Option<PathBuf>,
    env: HashMap<String, String>,
}

// QUESTION: Would String be less awkward?
impl From<&str> for UpdateSteps {
    fn from(value: &str) -> Self {
        match &*value.to_lowercase() {
            "fetch" => UpdateSteps::Fetch,
            "compile" => UpdateSteps::Compile,
            "install" => UpdateSteps::Install,

            _ => panic!("Step {} not recognized", value),
        }
    }
}

// QUESTION: Would String be less awkward?
impl From<&String> for UpdateSteps {
    fn from(value: &String) -> Self {
        match &*value.to_lowercase() {
            "fetch" => UpdateSteps::Fetch,
            "compile" => UpdateSteps::Compile,
            "install" => UpdateSteps::Install,

            _ => panic!("Step {} not recognized", value),
        }
    }
}

impl Updater {
    fn new() -> Updater {
        let mut up = Updater { systems: vec![] };
        up.systems.push(System {
            fetch: None,
            compile: None,
            install: Cmd::new(),
        });
        up
    }

    /// To create a sample config from code
    #[doc(hidden)]
    fn write_config(&self, opt: &Opt) {
        use std::fs::OpenOptions;

        let mut f = OpenOptions::new()
            .write(true)
            .create(true)
            .truncate(true)
            .open(&opt.config_file)
            .unwrap();

        f.write_all(serde_yaml::to_string(&self).unwrap().as_bytes())
            .unwrap();
    }

    pub fn from_config(opt: &Opt) -> Result<Updater> {
        // let u = Updater::new();
        // u.write_config(opt);

        let file = std::fs::read_to_string(&opt.config_file).unwrap();
        Ok(serde_yaml::from_str(&file).unwrap())
    }

    pub fn update_all(&self, opt: &Opt) -> Summary {
        // self.systems.iter().collect();
        let mut systems: Vec<_> = vec![];
        for sys in &self.systems {
            systems.push(self.update(&sys, opt));
        }
        Summary { systems }
    }

    fn update(&self, sys: &System, opt: &Opt) -> Result<()> {
        let steps = if opt.steps.is_empty() {
            vec![
                UpdateSteps::Fetch,
                UpdateSteps::Compile,
                UpdateSteps::Install,
            ]
        } else {
            opt.steps.iter().map(|u| u.into()).collect()
        };

        if steps.contains(&UpdateSteps::Fetch) {
            sys.fetch(opt)?;
        }
        if steps.contains(&UpdateSteps::Compile) {
            sys.compile(opt)?;
        }
        if steps.contains(&UpdateSteps::Install) {
            sys.install(opt)?;
        }

        Ok(())
    }
}

impl System {
    pub fn fetch(&self, opt: &Opt) -> Result<()> {
        if let Some(fetch) = &self.fetch {
            let cmd = fetch.clone().prepare(opt);
            cmd.execute(opt)
                .map_err(|err| MyError::new(ErrorKind::Fetch, err, cmd))?;
        }
        Ok(())
    }

    pub fn compile(&self, opt: &Opt) -> Result<()> {
        if let Some(compile) = &self.compile {
            let cmd = compile.clone().prepare(opt);
            cmd.execute(opt)
                .map_err(|err| MyError::new(ErrorKind::Compile, err, cmd))?;
        }
        Ok(())
    }

    pub fn install(&self, opt: &Opt) -> Result<()> {
        let cmd = self.install.clone().prepare(opt);
        cmd.execute(opt)
            .map_err(|err| MyError::new(ErrorKind::Install, err, cmd))?;
        Ok(())
    }
}

impl Cmd {
    fn new() -> Cmd {
        Cmd {
            exe: "".into(),
            params: vec![],
            current_dir: None,
            env: HashMap::new(),
        }
    }

    fn prepare(self, opt: &Opt) -> ActualCmd {
        let mut params = self.params;

        let mut env = self.env;
        if !env.contains_key("PATH") {
            env.insert("PATH".to_owned(), std::env!("PATH").to_owned());
        }

        ActualCmd {
            exe: self.exe,
            params,
            current_dir: self.current_dir,
            env,
        }
    }
}

impl ActualCmd {
    fn execute(&self, opt: &Opt) -> io::Result<ExitStatus> {
        let mut cmd = Command::new(&self.exe);

        cmd.args(&self.params).env_clear().envs(&self.env);

        if let Some(cdir) = &self.current_dir {
            cmd.current_dir(std::fs::canonicalize(cdir)?);
        }

        print!("Executing: {}", self);
        stdout().flush()?;

        if opt.quiet {
            // FIXME: stdin does not work with sudo?
            cmd.stdin(Stdio::null())
                .stdout(Stdio::null())
                .stderr(Stdio::null());
        }

        Ok(cmd.status()?)
    }
}

impl fmt::Display for ActualCmd {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let command = if !self.params.is_empty() {
            format!("{} {}", &self.exe, &self.params.join(" "))
        } else {
            self.exe.clone()
        };
        write!(f, "`{}`", command)?;

        if let Some(cdir) = &self.current_dir {
            write!(f, " in {:?}", cdir)?;
        }

        if !self.env.is_empty() {
            writeln!(f, " with the following environment variable:")?;
            writeln!(f, "{:#?}", self.env)
        } else {
            write!(f, " without any environment variable. ")
        }
    }
}
